/*
 This file is part of the digitalSTROM ds485-daemon (ds485d).

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-daemon is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-daemon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-daemon; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "config.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <syslog.h>
#include <time.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <ctype.h>

#include "utlist.h"
#include "uthash.h"

#include <digitalSTROM/ds485.h>
#include <digitalSTROM/ds485-core.h>
#include "ds485-socket-interface.h"


/******************************************************************************
 *
 */
static ds485Handle_t busHandle = NULL;

extern char *optarg;
extern int optind;

static int debugLevel = 5;
static int mainActive = 1;
static int gotSignal = 0;
static jmp_buf restartbuf;

static pthread_t rxThread;

/* Serial Device Name of the device we manage */
static char* serialDevice;

/* Some serial device (adapters) cannot receive own transmission data */
static int serialLoopback = 0;

/* process id file */
static char* pidfile = "/var/run/ds485d.pid";

/* server socket file descriptor */
static int sfd = -1;

/* listen on all network listen, or just loopback */
static int listenGlobal = 0;

/* running as daemon background process */
static int inBackground = 0;

/* configured dsid */
static dsid_t dsid = { {0x35, 0x04, 0x17, 0x5f, 0xe0, 0, 0, 0xf0, 0, 0, 0, 1} };

static char* loggerfile = NULL;

/* all client sockets */
static fd_set all_fds;
static int maxfd = -1;

struct transactionIdItem {
    int transactionId;
    struct transactionIdItem *next;
};
/** List of free TransactionIds */
static struct transactionIdItem *freeTransactionIds = NULL;

/**
 * Client session
 */
typedef struct _dS485dClient {
    struct _dS485dClient* next;

    int fd;                    /* client socket descriptor */
    int transactionId;        /* transactionId of the client */
    pthread_t tid;            /* thread id */

    ds485Handle_t* busHandle;    /* ds485 lib handle */

    uint8_t apirev;            /* client API revision */
    uint8_t rxfilter;        /* reception packet filter */
    uint8_t rawmode;        /* raw mode */

    double active;            /* last activity timestamp */
} dS485dClient;


/**
 * Device data
 */
typedef struct _dS485dDevice {
    UT_hash_handle hh;        /* hash handle */

    dsid_t dsid;            /* 96 Bit ID */
    dS485dClient* client;    /* corresponding client data */
} dS485dDevice;


static dS485dClient* clSessions = NULL;

static pthread_mutex_t clientListMutex;
static pthread_mutex_t reportMutex;
static pthread_mutex_t transactionIdListMutex;


/* Name of (unprivileged) user to change to when we drop privileges. */
#ifndef DS485D_USER
#define DS485D_USER    "nobody"
#endif

/*
 * Timeout policy for client connections
 */
#define NOREAD_TIMEOUT        60*3


/********************* SIGNAL & ERROR HANDLING *********************/

/*
 * Get the signal and save for main-loop processing
 */
void onsignal(int signo)
{
    gotSignal = signo;
    mainActive = 0;

    /* shutdown socket interface */
    close(sfd);
}

/**
 * Retrieve the next free TransactionId.
 *
 * @return the TransactionId; -1 if no TransactionId is left.
 */
static int getNextTransactionId(void)
{
    pthread_mutex_lock(&transactionIdListMutex);
    if (freeTransactionIds == NULL)
    {
        pthread_mutex_unlock(&transactionIdListMutex);
        return -1;
    }

    struct transactionIdItem *item = freeTransactionIds;
    int transactionId = item->transactionId;

    LL_DELETE(freeTransactionIds, item);
    free(item);

    pthread_mutex_unlock(&transactionIdListMutex);
    return transactionId;
}

/**
 * Return an unused TransactionId to the list of free TransactionIds.
 *
 * @param transactionId [in] the TranscationId
 * @return 0 on success; -1 on failure (out of memory)
 */
static int returnTransactionId(int transactionId)
{
    struct transactionIdItem *item = (struct transactionIdItem *)malloc(sizeof(struct transactionIdItem));
    if (item == NULL)
    {
        return -1;
    }
    item->transactionId = transactionId;
    pthread_mutex_lock(&transactionIdListMutex);
    LL_PREPEND(freeTransactionIds, item);
    pthread_mutex_unlock(&transactionIdListMutex);

    return 0;
}

/**
 * Initialize the list of available TransactionIds.
 *
 * @return 0 on success; -1 on failure (out of memory)
 */
static int initializeTransactionIdList(void)
{
    int id;
    for (id = 255; id != 0; --id)
    {
        if (returnTransactionId(id) < 0)
        {
            return -1;
        }
    }
    return 0;
}

static void deleteTransactionIdList(void)
{
    /* checking out all the TransactionIds will also free the allocated memory. */
    while (getNextTransactionId() >= 0) {}
}

static int daemonize(void)
{
    int fd;
    pid_t pid;

    switch (pid = fork()) {
    case -1:
        return -1;
    case 0:        /* child side */
        break;
    default:    /* parent side */
        exit(0);
    }

    if (setsid() == -1)
        return -1;
    if (chdir("/") == -1)
        return -1;

    if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
        dup2(fd, STDIN_FILENO);
        dup2(fd, STDOUT_FILENO);
        dup2(fd, STDERR_FILENO);
        if (fd > 2)
            close(fd);
    }
    inBackground = 1;
    return 0;
}

static void print_usage(void)
{
    fprintf(stderr, "Usage: dS485d [OPTION] <device>\n");
    fprintf(stderr, "       <device> is a tty name\n\n");
    fprintf(stderr, "       OPTION is a selection out of the following options:\n");
    fprintf(stderr, "       -h  --  print this help\n");
    fprintf(stderr, "       -d  --  debug-level: max debug level still logged: 0-7 (default: 5)\n");
    fprintf(stderr, "       -p  --  PID-file (default: /var/run/ds485d.pid)\n");
    fprintf(stderr, "       -F  --  stay in foreground (default: not set == daemonize)\n");
    fprintf(stderr, "               log-output: foreground: stdout\n");
    fprintf(stderr, "                           daemonized: syslog\n");
    fprintf(stderr, "       -G  --  enable external access: accept socket connection from other machines\n");
    fprintf(stderr, "       -L  --  enable extra serial loopback on RS485 interface\n");
    fprintf(stderr, "       -V  --  print ds485d version information\n");
    fprintf(stderr, "       -l  --  logger file: log raw data to this file.\n");
    fprintf(stderr, "                            Attention: access to this file must never block.\n");
    fprintf(stderr, "       -i  --  change dSID of ds485d (default: 3504175fe00000f000000001\n");
    fprintf(stderr, "\n");
}

static double timestamp(void) 
{
    struct timeval tv; 

    gettimeofday(&tv, NULL); 
    return (tv.tv_sec + tv.tv_usec*1e-6);
}

static void report(int errlevel, const char *fmt, ... )
{
    if (errlevel <= debugLevel) {
        char buf[BUFSIZ];
        va_list ap;
        pthread_mutex_lock(&reportMutex);

        strncpy(buf, "ds485d: ", BUFSIZ);
        va_start(ap, fmt) ;
        vsnprintf(buf + strlen(buf), sizeof(buf)-strlen(buf), fmt, ap);
        va_end(ap);

        if (inBackground) {
            syslog(errlevel, "%s", buf);
        }
        else {
            char buf2[BUFSIZ], *sp, tsbuf[30];
            struct timeval t;
            gettimeofday(&t, NULL);
            strftime(tsbuf, sizeof(tsbuf), "%Y-%m-%d %H:%M:%S", localtime(&t.tv_sec));
            (void)snprintf(buf2, BUFSIZ, "[%s.%03d] ", tsbuf, (int)t.tv_usec / 1000);
            for (sp = buf; *sp != '\0'; sp++)
                if (isprint(*sp) || (isspace(*sp) && (sp[1]=='\0' || sp[2]=='\0')))
                    (void)snprintf(buf2+strlen(buf2), 2, "%c", *sp);
                else
                    (void)snprintf(buf2+strlen(buf2), 6, "\\x%02x", (unsigned)*sp);
            (void)fputs(buf2, stderr);
        }

        pthread_mutex_unlock(&reportMutex);
    }
}

static int passivesock(char *service, char *protocol, int qlen)
{
    struct servent *pse;
    struct protoent *ppe;
    struct sockaddr_in sin;
    int s, type, proto, one = 1;

    memset((char *) &sin, 0, sizeof(sin));
    sin.sin_family = (sa_family_t)AF_INET;
    if (listenGlobal)
        sin.sin_addr.s_addr = htonl(INADDR_ANY);
    else
        sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

    if ((pse = getservbyname(service, protocol)) != NULL)
        sin.sin_port = htons(ntohs((in_port_t)pse->s_port));
    else if ((sin.sin_port = htons((in_port_t)atoi(service))) == 0) {
        report(LOG_INFO, "Can't get \"%s\" service entry, using default port 8442.\n",
               service);
        sin.sin_port = htons(8442);
    }

    ppe = getprotobyname(protocol);
    if (strcmp(protocol, "udp") == 0) {
        type = SOCK_DGRAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_UDP;
    } else {
        type = SOCK_STREAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_TCP;
    }
    if ((s = socket(PF_INET, type, proto)) < 0) {
        report(LOG_ERR, "Can't create socket\n");
        return -1;
    }
    if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,(int)sizeof(one)) == -1) {
        report(LOG_ERR, "Error: SETSOCKOPT SO_REUSEADDR\n");
        return -1;
    }
    if (bind(s, (struct sockaddr *) &sin, (int)sizeof(sin)) < 0) {
        report(LOG_ERR, "Can't bind to port %s\n", service);
        if (errno == EADDRINUSE) {
            report(LOG_ERR, "Maybe ds485d is already running!\n");
        }
        return -1;
    }
    if (type == SOCK_STREAM && listen(s, qlen) < 0) {
        report(LOG_ERR, "Can't listen on port %s\n", service);
        return -1;
    }

    int opts = fcntl(s, F_GETFL);
    if (opts > 0) {
        fcntl(s, F_SETFL, opts | O_NONBLOCK);
    }
    #ifdef IPTOS_LOWDELAY
    {
        int opt = IPTOS_LOWDELAY;
        setsockopt(s, IPPROTO_IP, IP_TOS, &opt, sizeof opt);
    }
    #endif
    #ifdef TCP_NODELAY
    {
        if (type == SOCK_STREAM)
            setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &one, sizeof one);
    }
    #endif

    return s;
}

static int filesock(char *filename)
{
    struct sockaddr_un addr;
    int sock;

    if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
        report(LOG_ERR, "Can't create control socket\n");
        return -1;
    }
    strncpy(addr.sun_path, filename, 104);
    addr.sun_family = AF_UNIX;
    bind(sock, (struct sockaddr *) &addr,  (socklen_t)sizeof(addr));
    if (listen(sock, 5) < 0) {
        report(LOG_ERR, "Can't listen on local socket %s\n", filename);
        return -1;
    }
    return sock;
}

char *sock2ip(int fd)
{
    struct sockaddr fsin;
    socklen_t alen = (socklen_t)sizeof(fsin);
    char *ip;
    int r;

    r = getpeername(fd, (struct sockaddr *) &fsin, &alen);
    if (r == 0) {
        ip = inet_ntoa(((struct sockaddr_in *)(&fsin))->sin_addr);
    } else {
        report(LOG_INFO, "getpeername() = %d, error = %s (%d)\n", r, strerror(errno), errno);
        ip = "<unknown>";
    }
    return ip;
}

/* track the largest fd currently in use */
static void clAdjustMax(int fd, int on)
{
    if (on) {
        if (fd > maxfd)
            maxfd = fd;
    }
    else {
        if (fd == maxfd) {
            int tfd;

            for (maxfd = tfd = 0; tfd < FD_SETSIZE; tfd++)
                if (FD_ISSET(tfd, &all_fds))
                    maxfd = tfd;
        }
    }
}

static void clDetach(dS485dClient* cl)
{
    char *c_ip;

    if (-1 == cl->fd)
        return;

    c_ip = sock2ip(cl->fd);
    shutdown(cl->fd, SHUT_RDWR);
    close(cl->fd);

    report(LOG_INFO, "detaching %s (connection %d, fd %d)\n",
           c_ip, cl->transactionId, cl->fd);
    FD_CLR(cl->fd, &all_fds);

    clAdjustMax(cl->fd, 0);

    cl->rawmode = 0;
    cl->active = 0;
    cl->fd = -1;
}

static int clWrite(dS485dClient* cl, uint8_t* data, size_t dlen)
{
    int status = write(cl->fd, data, dlen);

    if (status == dlen ) {
        return status;
    }
    else if (status > -1) {
        report(LOG_INFO, "short write disconnecting client(%d)\n", cl->transactionId);
        clDetach(cl);
        return 0;
    } else if (errno == EAGAIN || errno == EINTR) {
        /* no data written, try again */
        return 0;
    }
    else if (errno == EBADF) {
        report(LOG_WARNING, "client(%d) has vanished.\n", cl->transactionId);
    }
    else if (errno == EWOULDBLOCK && (timestamp() - cl->active) > NOREAD_TIMEOUT) {
        report(LOG_INFO, "client(%d) timed out.\n", cl->transactionId);
    }
    else {
        report(LOG_INFO, "client(%d) write: %s\n", cl->transactionId, strerror(errno));
    }
    clDetach(cl);
    return status;
}

static int clWriteCommand(dS485dClient* cl, uint8_t cmd, uint8_t* data, int dlen)
{
    uint8_t buf[8];
    uint16_t length = dlen + 3;

    buf[0] = cmd;
    buf[1] = length & 0xff;
    buf[2] = (length >> 8) & 0xff;

    int rval = clWrite(cl, buf, 3);
    if (rval != 3)
        return rval;

    if (data && dlen > 0 ) {
        rval = clWrite(cl, data, dlen);
        if (rval != dlen)
            return rval;
    }

    return 0;
}


/*************************** Protocol Callbacks *******************************/
void callback_bus_change(ds485Handle_t dev, dsid_t *dsid, int flags)
{
    dS485dClient* cs;
    ds485_container_t container;
    int n, containerLength;
    char buffer[64];
    char *p;

    p = buffer;
    for (n = 0; n < sizeof(dsid_t); n++)
        p += sprintf(p, "%02x", dsid->id[n]);
    *p = 0;

    report(LOG_NOTICE, "bus change: %s -> %i\n", buffer, flags);

    container.sourceId = *dsid;
    container.data[0] = flags;
    container.length = 1;

    containerLength = DS485_CONTAINER_SIZE + container.length;

    /* deliver to each client session */
    pthread_mutex_lock(&clientListMutex);
    LL_FOREACH(clSessions, cs) {
        if (cs->fd >= 0) {
            clWriteCommand(cs,
                           DS485D_CMD_BUS_CHANGE,
                           (uint8_t*) &container, containerLength);
        }
    }
    pthread_mutex_unlock(&clientListMutex);
}

/**
 * registered system callback for protocol (debug) messages
 */
int callback_system_msg(ds485Handle_t dev, unsigned char level, const char *msg, va_list args)
{
    char buffer[512];

    vsnprintf(buffer, 512, msg, args);

    report(level, "%s", buffer);
    return 0;
}

/**
 * registered device callback for bus state changes
 */
int callback_state(ds485Handle_t dev, bus_state_t state)
{
    dS485dClient* cs;
    ds485_container_t container;
    int containerLength;

    report(LOG_NOTICE, "bus state change: %i\n", state );

    container.data[0] = state;
    container.length = 1;

    containerLength = DS485_CONTAINER_SIZE + container.length;

    /* deliver to each client session */
    pthread_mutex_lock(&clientListMutex);
    LL_FOREACH(clSessions, cs) {
        if (cs->fd >= 0) {
            clWriteCommand(cs,
                           DS485D_CMD_BUS_STATE,
                           (uint8_t*) &container, containerLength);
        }
    }
    pthread_mutex_unlock(&clientListMutex);
    
    return 0;
}

/**
 * registered device callback for frame reception
 */
int callback_frame(ds485Handle_t dev, ds485_container_t const * frame)
{
    if (NULL == dev)
        return -1;

#if 0
    report(LOG_DEBUG, "Device %02d: frame ...\n",
        dev->busAddress);
#endif

    return 0;
}


/******************************* RECEIVE THREAD *******************************/
static uint64_t rxPacketNum = 0;
static uint64_t txPacketNum = 0;

static void* sysRxThread(void* arg)
{
    ds485Handle_t *handle = (ds485Handle_t *) arg;
    dS485dClient* cs;
    ds485_container_t container;
    unsigned int containerLength;
    int rval;

    /* Read incoming ds485 data */
    while (mainActive) {

        rval = ds485_api_recv_data(handle, &container);

        if (rval < 0) {
            continue;
        }
        rxPacketNum++;

        /* deliver to each client session */
        pthread_mutex_lock(&clientListMutex);
        LL_FOREACH(clSessions, cs) {
            if (cs->fd >= 0 &&
               ((cs->transactionId == container.transactionId) ||
                (container.containerType != DS485_CONTAINER_RESPONSE) ||
                ((cs->rxfilter & DS485_FILTER_ALL_RESPONSES) == DS485_FILTER_ALL_RESPONSES))) {

                if (cs->rawmode) {
                    rval = clWriteCommand(cs,
                            DS485D_CMD_DEVICE_PACKET,
                            (uint8_t*) &container, sizeof(ds485_container_t));
                }
                else {
                    containerLength = DS485_CONTAINER_SIZE + container.length;

                    rval = clWriteCommand(cs,
                            DS485D_CMD_RX_PACKET,
                            (uint8_t*) &container, containerLength);
                }

                txPacketNum++;
            }
        }
        pthread_mutex_unlock(&clientListMutex);

    }

    return NULL;
}

/******************************* SERVICE THREAD *******************************/

static int clSend(dS485dClient *cl, ds485_container_t* pContainer, uint16_t length)
{
    uint16_t clength = pContainer->length;

    if (length < DS485_CONTAINER_SIZE + clength) {
        report(LOG_ERR, "client container data has too less data, ignore %d byte\n", clength);
        return -1;
    }

    if (clength > DS485_MAX_PAYLOAD_SIZE) {
        report(LOG_ERR, "client container data too large, ignore %d byte\n", clength);
        return (clength + DS485_CONTAINER_SIZE);
    }

    pContainer->transactionId = cl->transactionId;
    pContainer->sourceId = dsid;

    ds485_api_send_data(cl->busHandle, pContainer);

    /* all packets have to be sent to all clients except the sender, too. */
    dS485dClient* cs;
    int rval;

    pthread_mutex_lock(&clientListMutex);
    LL_FOREACH(clSessions, cs) {
        if ((cs->fd >= 0) &&
            ((cs != cl) || ((cs->rxfilter & DS485_FILTER_ALL_RESPONSES) == DS485_FILTER_ALL_RESPONSES))) {

            if (cs->rawmode) {
                rval = clWriteCommand(cs,
                                      DS485D_CMD_DEVICE_PACKET,
                                      (uint8_t*) pContainer, sizeof(ds485_container_t));
            }
            else {
                int containerLength = DS485_CONTAINER_SIZE + pContainer->length;

                rval = clWriteCommand(cs,
                                      DS485D_CMD_RX_PACKET,
                                      (uint8_t*) pContainer, containerLength);
            }

            txPacketNum++;
        }
    }
    pthread_mutex_unlock(&clientListMutex);

    return (clength + DS485_CONTAINER_SIZE);
}

static int sysServiceDispatch(dS485dClient *cl, uint8_t* data, int dlen)
{
    dS485dCommand* cmd;
    uint8_t value;
    uint16_t len;
    int rval;

    cmd = (dS485dCommand *) data;
    len = cmd->length;

    cl->active = timestamp();
    report(LOG_DEBUG, "client command: 0x%02x, len 0x%02x, data length %d\n", cmd->command, len, dlen);

    switch (cmd->command) {

    case DS485D_CMD_API_REVISION:
        cl->apirev = cmd->data[0];
        value = DS485D_API_REVISION;
        clWriteCommand(cl, DS485D_CMD_API_REVISION, &value, 1);
        break;

    case DS485D_CMD_PING:
        clWriteCommand(cl, DS485D_CMD_PING, NULL, 0);
        break;

    case DS485D_CMD_TX_PACKET: {
        dS485CmdApidata *apiData = (dS485CmdApidata *) cmd;
        rval = clSend(cl, &apiData->apiData, apiData->length - 3);
        if (rval > 0) {
            /* only remove the processed data bytes in the first container */
            len = rval + 3;
        }
        break;
    }
    case DS485D_CMD_RX_PACKET:
        report(LOG_WARNING, "invalid client command\n");
        break;

    case DS485D_CMD_DEVICE_PACKET:
        if (cl->rawmode)
        {
            ds485_container_t* packet = (ds485_container_t *) &cmd->data[0];
            ds485_api_send_data(cl->busHandle, packet);
            len = sizeof(ds485_container_t) + 3;
        }
        else {
            report(LOG_WARNING, "client tried to send raw packet, rawmode is not enabled\n");
        }
        break;

    case DS485D_CMD_PACKETFILTER:
        cl->rxfilter = cmd->data[0];
        report(LOG_DEBUG, "client %d: set rxfilter: 0x%02x\n",
               cl->transactionId, cl->rxfilter);
        break;

    case DS485D_CMD_RAWMODE:
        cl->rawmode = cmd->data[0];
        report(LOG_DEBUG, "client %d: rawmode %s\n",
               cl->transactionId, cl->rawmode ? "enable" : "disable");
        break;

    case DS485D_CMD_QUERY_DEVICES:
    {
        int length = cmd->data[0];
        dsid_t *devices = (dsid_t*)malloc(length * sizeof(dsid_t));
        length = ds485_api_get_bus_members(cl->busHandle, devices, length);
        clWriteCommand(cl, DS485D_CMD_QUERY_DEVICES, (uint8_t *)devices, length * sizeof(dsid_t));
        free(devices);
        break;
    }

    case DS485D_CMD_QUERY_DSID:
    {
        clWriteCommand(cl, DS485D_CMD_QUERY_DSID, (uint8_t *)&dsid, sizeof(dsid_t));
        break;
    }

    case DS485D_CMD_QUERY_ADDRESS:
        /* TODO: send reply */
        break;

    case DS485D_CMD_QUERY_BUSSTATE:
        cmd->data[0] = ds485_api_get_bus_state(cl->busHandle);
        clWriteCommand(cl, DS485D_CMD_QUERY_BUSSTATE, &cmd->data[0], 1);
        break;

    default:
        report(LOG_WARNING, "unknown client command\n");
        break;
    }

    return len;
}

static void* sysServiceThread(void* arg)
{
    dS485dClient *cl = (dS485dClient *) arg;

    struct timeval tv;
    fd_set rfds;

    int rval;
    int rxLen = 0;

#define SVCBUFLEN 1024
    uint8_t* rxBuffer = (uint8_t *) malloc(SVCBUFLEN);
    uint8_t* p;

    char *c_ip = sock2ip(cl->fd);
    report(LOG_NOTICE, "new client connection %d from %s\n", cl->transactionId, c_ip);

    /* handle client connection */
    while (mainActive && cl->fd > 0) {

        FD_ZERO(&rfds);
        FD_SET(cl->fd, &rfds);

        tv.tv_sec = 1;
        tv.tv_usec = 0;
        rval = select(cl->fd + 1, &rfds, NULL, NULL, &tv);

        if (rval < 0) {
            if (errno == EINTR)
                continue;
            report(LOG_WARNING, "client error %d [%d]\n", rval, cl->fd);
            clDetach(cl);
            break;
        }
        else if (rval == 0) {
            /* timeout */
            if (rxLen != 0) {
                report(LOG_WARNING, "client timeout: reset buffer (prev: %d) [%d]\n", rxLen, cl->fd);
            }
            rxLen = 0;
            continue;
        }
        if (FD_ISSET(cl->fd, &rfds)) {
            rval = recv(cl->fd, rxBuffer + rxLen, SVCBUFLEN - rxLen, 0);
            if (rval < 0) {
                report(LOG_WARNING, "client read error %d [%d]\n", rval, cl->fd);
                continue;
            }
            else if (rval == 0) {
                clDetach(cl);
                break;
            }
            else {
                rxLen += rval;
            }
        }
        else {
            continue;
        }

        p = rxBuffer;
        while (rxLen >= 3) {

            /* sanity checks */
            if (p[0] == 0 && p[1] == 0 && p[2] == 0) {
                rxLen = 0;
                report(LOG_INFO, "client resync %s, fd %d\n", c_ip, cl->fd);
                break;
            }

            /* not enough data */
            if ((p[2] << 8 | p[1]) > rxLen) {
                report(LOG_INFO, "client has not enough data, read again (%d > %d)\n", (p[2] << 8 | p[1]), rxLen);
                break;
            }

            /* Process packet data ... */
            rval = sysServiceDispatch(cl, p, rxLen);

            p += rval;
            rxLen -= rval;

        }
        if (p != rxBuffer && rxLen > 0) {
            memcpy(rxBuffer, p, rxLen);
        }

    }
    report(LOG_NOTICE, "terminated connection %d from %s\n", cl->transactionId, c_ip);

    if (returnTransactionId(cl->transactionId) < 0)
    {
        report(LOG_ERR, "could not add transactionId: %d to the free list\n", cl->transactionId);
    }
    /*
     * clDetach is already called, socket is closed
     */
    pthread_mutex_lock(&clientListMutex);
    LL_DELETE(clSessions, cl);
    pthread_mutex_unlock(&clientListMutex);

    free(cl);
    free(rxBuffer);
    return NULL;
}


/******************************* MAIN *******************************/

/*
 * Function main (argc, )
 *
 *    Main function
 *
 */

int main(int argc, char* argv[])
{
    int c;
    int goBackground = 1;

    static int st;
    struct sockaddr fsin;
    fd_set sfds;
    struct timeval tv;

    debugLevel = 5;
    mainActive = 1;

    pthread_mutex_init(&clientListMutex, NULL);
    pthread_mutex_init(&reportMutex, NULL);
    pthread_mutex_init(&transactionIdListMutex, NULL);

#ifdef HAVE_SETLOCALE
    setlocale(LC_NUMERIC, "C");
#endif

    if (argc < 2) {
        print_usage();
        exit(-1);
    }

    /* Look for options */
    while ((c = getopt(argc, argv, "hd:p:FGLVl:i:")) != -1) {
        switch (c) {
        case 'h':
            print_usage();
            exit(0);
        case 'd':
            debugLevel = atoi(optarg);
            break;
        case 'p':
            pidfile = strdup(optarg);
            break;
        case 'F':
            goBackground = 0;
            break;
        case 'G':
            listenGlobal = 1;
            break;
        case 'L':
            serialLoopback = 1;
            break;
        case 'l':
            loggerfile = strdup(optarg);
            break;
        case 'i':
        {
            const char *dsid_str = optarg;

            if (strlen(dsid_str) != 24)
            {
                report(LOG_ERR, "dSID in the wrong format\n");
                exit(-1);
            }
            sscanf(dsid_str,
                   "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
                   (unsigned int *)&dsid.id[0],
                   (unsigned int *)&dsid.id[1],
                   (unsigned int *)&dsid.id[2],
                   (unsigned int *)&dsid.id[3],
                   (unsigned int *)&dsid.id[4],
                   (unsigned int *)&dsid.id[5],
                   (unsigned int *)&dsid.id[6],
                   (unsigned int *)&dsid.id[7],
                   (unsigned int *)&dsid.id[8],
                   (unsigned int *)&dsid.id[9],
                   (unsigned int *)&dsid.id[10],
                   (unsigned int *)&dsid.id[11]);
            break;
        }
        case 'V':
            printf("Version: v%s\n", VERSION);
            exit(0);
        default:
            exit(-1);
        }
    }

    /* Last arg is device or file name. */
    serialDevice = strdup(argv[argc-1]);

    if (initializeTransactionIdList() < 0)
    {
        report(LOG_ERR, "Initialization of list of free transactionIds failed\n");
        exit(-1);
    }

    /* go to background mode */
    if (goBackground) {
        openlog("ds485d", LOG_PID, LOG_USER);
        daemonize();
    }
    report(LOG_NOTICE, "%s, %s (built %s, %s)\n", PROJECTNAME, VERSION, __DATE__, __TIME__);

    /* PID file */
    if (pidfile) {
        FILE* fp;

        if ((fp = fopen(pidfile, "w")) != NULL) {
            fprintf(fp, "%u\n", (unsigned int) getpid());
            fclose(fp);
        } else {
            report(LOG_ERR, "Cannot create PID file: %s\n", pidfile);
        }
    }

    /* got hangup signal */
    if ((st = setjmp(restartbuf)) > 0) {
        report(LOG_WARNING, "ds485d restarted by SIGHUP\n");
    }

    /* Trap signals */
    gotSignal = 0;
    signal(SIGINT, onsignal);
    signal(SIGTERM, onsignal);
    signal(SIGQUIT, onsignal);
    signal(SIGHUP, onsignal);
    signal(SIGPIPE, onsignal);

    /* start ds485 protocol */
    busHandle = ds485_api_open(serialDevice, &dsid);
    if (NULL == busHandle) {
        report(LOG_ERR, "ERROR initializing dS485 protocol\n");
        exit(2);
    }

    /* register message handler */
    ds485_api_ioctl(busHandle, IOC_SET_MSG_CB,
                    callback_system_msg, sizeof(callback_system_msg));

    ds485_api_ioctl(busHandle, IOC_SET_BUS_CHANGE_CB,
                    callback_bus_change, sizeof(callback_bus_change));

    ds485_api_ioctl(busHandle, IOC_SET_BUS_STATE_CB,
                    callback_state, sizeof(callback_state));

    /* configure software loopback */
    if (serialLoopback) {
        ds485_api_ioctl(busHandle, IOC_LOOPBACK_MODE,
                        &serialLoopback, sizeof(serialLoopback));
    }

    if (loggerfile)
    {
        int logger = open(loggerfile, O_WRONLY | O_NONBLOCK);
        if (logger < 0)
        {
            report(LOG_ERR, "Could not open logger file.\n");
        }
        else
        {
            report(LOG_INFO, "Logging raw data to: %s\n", loggerfile);
            ds485_api_ioctl(busHandle, IOC_SET_LOGGER_FD, &logger, sizeof(logger));
        }
    }

    /* Start receiver thread */
    if (pthread_create(&rxThread, NULL, sysRxThread, busHandle)) {
        report(LOG_ERR, "Could not create serial thread.\n");
        mainActive = 0;
    }
    pthread_detach(rxThread);

    /* Initialize server socket */
    sfd = passivesock("ds485", "tcp", 0);
    if (sfd < 0) {
        report(LOG_ERR, "ERROR opening network socket\n");
        exit(2);
    }
    FD_ZERO(&sfds);
    FD_SET(sfd, &sfds);

    /* Client connection loop */
    while (mainActive && !gotSignal) {
        socklen_t cnamelen;
        int cfd;

        /* wait for a client connection */
        tv.tv_sec = 1; tv.tv_usec = 0;
        if (select(1, &sfds, NULL, NULL, &tv) < 0) {
            if (errno == EINTR)
                continue;
        }

        /* new client */
        cnamelen = sizeof(fsin);
        cfd = accept(sfd, (struct sockaddr *) &fsin, &cnamelen);
        if (cfd <= 0) {
            continue;
        }

        int opts = fcntl(sfd, F_GETFL);
        if (opts > 0) {
            fcntl(cfd, F_SETFL, opts | O_NONBLOCK);
        }

        dS485dClient* cls = (dS485dClient*) malloc(sizeof(dS485dClient));
        cls->fd = cfd;
        cls->busHandle = busHandle;
        cls->rxfilter = 0;
        cls->transactionId = getNextTransactionId();
        if (cls->transactionId < 0)
        {
        	report(LOG_ERR, "no transaction id available\n");
            clDetach(cls);
            free(cls);
            continue;
        }
        cls->rawmode = 0;
        cls->active = timestamp();

        pthread_mutex_lock(&clientListMutex);
        LL_PREPEND(clSessions, cls);
        pthread_mutex_unlock(&clientListMutex);

        /* start service thread */
        if (pthread_create(&cls->tid, NULL, sysServiceThread, cls)) {
            report(LOG_ERR, "Could not create client thread.\n");
            mainActive = 0;
        }
        pthread_detach(cls->tid);
    }

    switch (gotSignal) {
    case SIGTERM:
    case SIGINT:
        report(LOG_NOTICE, "got SIGTERM or SIGINT\n");
        mainActive = 0;
        break;
    case SIGHUP:
        report(LOG_NOTICE, "got SIGHUP\n");
        longjmp(restartbuf, 1);
        break;
    default:
        break;
    }

    {
        /* close all connections */
        dS485dClient* cs;
        pthread_mutex_lock(&clientListMutex);
        LL_FOREACH(clSessions, cs) {
            clDetach(cs);
        }
        pthread_mutex_unlock(&clientListMutex);
    }

    /* Cleanup threads */
    pthread_cancel(rxThread);

    /* shutdown ds485 protocol */
    ds485_api_close(busHandle);

    /* cleanup TransactionIds */
    deleteTransactionIdList();

    return 0;
}
